use std::net::{Ipv6Addr, Ipv4Addr};
use std::fmt;

use serde::{Serialize, Deserialize};
use trust_dns_client::serialize::binary::BinEncoder;
use base64;

use super::trust_dns_types;

#[derive(Deserialize, Serialize)]
pub enum RecordType {
    A,
    AAAA,
    ANAME,
    ANY,
    AXFR,
    CAA,
    CNAME,
    HINFO,
    HTTPS,
    IXFR,
    MX,
    NAPTR,
    NS,
    NULL,
    OPENPGPKEY,
    OPT,
    PTR,
    SOA,
    SRV,
    SSHFP,
    SVCB,
    TLSA,
    TXT,
    // dnssec
    DNSKEY,
    DS,
    KEY,
    NSEC,
    NSEC3,
    NSEC3PARAM,
    RRSIG,
    SIG,

    Unknown(u16),
    ZERO,
}


impl From<trust_dns_types::RecordType> for RecordType {
    fn from(record_type: trust_dns_types::RecordType) -> RecordType {
        match record_type {
            trust_dns_types::RecordType::A => RecordType::A,
            trust_dns_types::RecordType::AAAA => RecordType::AAAA,
            trust_dns_types::RecordType::ANAME => RecordType::ANAME,
            trust_dns_types::RecordType::ANY => RecordType::ANY,
            trust_dns_types::RecordType::AXFR => RecordType::AXFR,
            trust_dns_types::RecordType::CAA => RecordType::CAA,
            trust_dns_types::RecordType::CNAME => RecordType::CNAME,
            trust_dns_types::RecordType::HINFO => RecordType::HINFO,
            trust_dns_types::RecordType::HTTPS => RecordType::HTTPS,
            trust_dns_types::RecordType::IXFR => RecordType::IXFR,
            trust_dns_types::RecordType::MX => RecordType::MX,
            trust_dns_types::RecordType::NAPTR => RecordType::NAPTR,
            trust_dns_types::RecordType::NS => RecordType::NS,
            trust_dns_types::RecordType::NULL => RecordType::NULL,
            trust_dns_types::RecordType::OPENPGPKEY => RecordType::OPENPGPKEY,
            trust_dns_types::RecordType::OPT => RecordType::OPT,
            trust_dns_types::RecordType::PTR => RecordType::PTR,
            trust_dns_types::RecordType::SOA => RecordType::SOA,
            trust_dns_types::RecordType::SRV => RecordType::SRV,
            trust_dns_types::RecordType::SSHFP => RecordType::SSHFP,
            trust_dns_types::RecordType::SVCB => RecordType::SVCB,
            trust_dns_types::RecordType::TLSA => RecordType::TLSA,
            trust_dns_types::RecordType::TXT => RecordType::TXT,

            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::DNSKEY) => RecordType::DNSKEY,
            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::DS) => RecordType::DS,
            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::KEY) => RecordType::KEY,
            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::NSEC) => RecordType::NSEC,
            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::NSEC3) => RecordType::NSEC3,
            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::NSEC3PARAM) => RecordType::NSEC3PARAM,
            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::RRSIG) => RecordType::RRSIG,
            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::SIG) => RecordType::SIG,
            trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::Unknown(r#type)) => RecordType::Unknown(r#type),

            trust_dns_types::RecordType::Unknown(r#type) => RecordType::Unknown(r#type),
            trust_dns_types::RecordType::ZERO => RecordType::ZERO,
        }
    }
}

impl Into<trust_dns_types::RecordType> for RecordType {
    fn into(self) -> trust_dns_types::RecordType {
        match self {
            RecordType::A => trust_dns_types::RecordType::A,
            RecordType::AAAA => trust_dns_types::RecordType::AAAA,
            RecordType::ANAME => trust_dns_types::RecordType::ANAME,
            RecordType::ANY => trust_dns_types::RecordType::ANY,
            RecordType::AXFR => trust_dns_types::RecordType::AXFR,
            RecordType::CAA => trust_dns_types::RecordType::CAA,
            RecordType::CNAME => trust_dns_types::RecordType::CNAME,
            RecordType::HINFO => trust_dns_types::RecordType::HINFO,
            RecordType::HTTPS => trust_dns_types::RecordType::HTTPS,
            RecordType::IXFR => trust_dns_types::RecordType::IXFR,
            RecordType::MX => trust_dns_types::RecordType::MX,
            RecordType::NAPTR => trust_dns_types::RecordType::NAPTR,
            RecordType::NS => trust_dns_types::RecordType::NS,
            RecordType::NULL => trust_dns_types::RecordType::NULL,
            RecordType::OPENPGPKEY => trust_dns_types::RecordType::OPENPGPKEY,
            RecordType::OPT => trust_dns_types::RecordType::OPT,
            RecordType::PTR => trust_dns_types::RecordType::PTR,
            RecordType::SOA => trust_dns_types::RecordType::SOA,
            RecordType::SRV => trust_dns_types::RecordType::SRV,
            RecordType::SSHFP => trust_dns_types::RecordType::SSHFP,
            RecordType::SVCB => trust_dns_types::RecordType::SVCB,
            RecordType::TLSA => trust_dns_types::RecordType::TLSA,
            RecordType::TXT => trust_dns_types::RecordType::TXT,

            RecordType::DNSKEY => trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::DNSKEY),
            RecordType::DS => trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::DS),
            RecordType::KEY => trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::KEY),
            RecordType::NSEC => trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::NSEC),
            RecordType::NSEC3 => trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::NSEC3),
            RecordType::NSEC3PARAM => trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::NSEC3PARAM),
            RecordType::RRSIG => trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::RRSIG),
            RecordType::SIG => trust_dns_types::RecordType::DNSSEC(trust_dns_types::DNSSECRecordType::SIG),

            RecordType::Unknown(r#type) => trust_dns_types::RecordType::Unknown(r#type),
            RecordType::ZERO => trust_dns_types::RecordType::ZERO,
        }
    }
}


#[derive(Deserialize, Serialize)]
#[serde(tag = "Type")]
#[serde(rename_all = "UPPERCASE")]
pub enum RData {
    #[serde(rename_all = "PascalCase")]
    A {
        address: Ipv4Addr
    },
    #[serde(rename_all = "PascalCase")]
    AAAA {
        address: Ipv6Addr
    },
    #[serde(rename_all = "PascalCase")]
    CAA {
        issuer_critical: bool,
        value: String,
        property_tag: String,
    },
    #[serde(rename_all = "PascalCase")]
    CNAME {
        target: String
    },
    // HINFO(HINFO),
    // HTTPS(SVCB),
    #[serde(rename_all = "PascalCase")]
    MX {
        preference: u16,
        mail_exchanger: String
    },
    // NAPTR(NAPTR),
    #[serde(rename_all = "PascalCase")]
    NULL {
        data: String
    },
    #[serde(rename_all = "PascalCase")]
    NS {
        target: String
    },
    // OPENPGPKEY(OPENPGPKEY),
    // OPT(OPT),
    #[serde(rename_all = "PascalCase")]
    PTR {
        target: String
    },
    #[serde(rename_all = "PascalCase")]
    SOA {
        master_server_name: String,
        maintainer_name: String,
        refresh: i32,
        retry: i32,
        expire: i32,
        minimum: u32,
        serial: u32
    },
    // SRV(SRV),
    // SSHFP(SSHFP),
    // SVCB(SVCB),
    // TLSA(TLSA),
    // TXT(TXT),

    // TODO: Eventually allow deserialization of DNSSEC records
    #[serde(skip)]
    DNSSEC(trust_dns_types::DNSSECRData),
    #[serde(rename_all = "PascalCase")]
    Unknown {
        code: u16,
        data: String,
    },
    // ZERO,
}

impl From<trust_dns_types::RData> for RData {
    fn from(rdata: trust_dns_types::RData) -> RData {
        match rdata {
            trust_dns_types::RData::A(address) => RData::A { address },
            trust_dns_types::RData::AAAA(address) => RData::AAAA { address },
            // Still a draft, no iana number yet, I don't to put something that is not currently supported so that's why NULL and not unknown.
            // TODO: probably need better error here, I don't know what to do about that as this would require to change the From for something else.
            // (empty data because I'm lazy)
            trust_dns_types::RData::ANAME(_) =>  RData::NULL {
                data: String::new()
            },
            trust_dns_types::RData::CNAME(target) => RData::CNAME {
                target: target.to_utf8()
            },
            trust_dns_types::RData::CAA(caa) => RData::CAA {
                issuer_critical: caa.issuer_critical(),
                value: format!("{}", CAAValue(caa.value())),
                property_tag: caa.tag().as_str().to_string(),
            },
            trust_dns_types::RData::MX(mx) => RData::MX {
                preference: mx.preference(),
                mail_exchanger: mx.exchange().to_utf8()
            },
            trust_dns_types::RData::NULL(null) =>  RData::NULL {
                data: base64::encode(null.anything().map(|data| data.to_vec()).unwrap_or_default())
            },
            trust_dns_types::RData::NS(target) => RData::NS {
                target: target.to_utf8()
            },
            trust_dns_types::RData::PTR(target) => RData::PTR {
                target: target.to_utf8()
            },
            trust_dns_types::RData::SOA(soa) => RData::SOA {
                master_server_name: soa.mname().to_utf8(),
                maintainer_name: soa.rname().to_utf8(),
                refresh: soa.refresh(),
                retry: soa.retry(),
                expire: soa.expire(),
                minimum: soa.minimum(),
                serial: soa.serial()
            },
            trust_dns_types::RData::DNSSEC(data) => RData::DNSSEC(data),
            rdata => {
                let code = rdata.to_record_type().into();
                let mut data = Vec::new();
                let mut encoder = BinEncoder::new(&mut data);
                rdata.emit(&mut encoder).expect("could not encode data");

                RData::Unknown {
                    code,
                    data: base64::encode(data),
                }
            }
        }
    }
}

struct CAAValue<'a>(&'a trust_dns_types::caa::Value);

// trust_dns Display implementation panics if no parameters
// Implementation based on caa::emit_value
// Also the quotes are strips to render in JSON
impl<'a> fmt::Display for CAAValue<'a> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
        match self.0 {
            trust_dns_types::caa::Value::Issuer(name, parameters) => {

                if let Some(name) = name {
                    write!(f, "{}", name)?;
                }

                if name.is_none() && parameters.is_empty() {
                    write!(f, ";")?;
                }

                for value in parameters {
                    write!(f, "; {}", value)?;
                }
            }
            trust_dns_types::caa::Value::Url(url) => write!(f, "{}", url)?,
            trust_dns_types::caa::Value::Unknown(v) => write!(f, "{:?}", v)?,
        }
        Ok(())
    }
}

#[derive(Deserialize, Serialize)]
pub enum DNSClass {
    IN,
    CH,
    HS,
    NONE,
    ANY,
    OPT(u16),
}

impl From<trust_dns_types::DNSClass> for DNSClass {
    fn from(dns_class: trust_dns_types::DNSClass) -> DNSClass {
        match dns_class {
            trust_dns_types::DNSClass::IN => DNSClass::IN,
            trust_dns_types::DNSClass::CH => DNSClass::CH,
            trust_dns_types::DNSClass::HS => DNSClass::HS,
            trust_dns_types::DNSClass::NONE => DNSClass::NONE,
            trust_dns_types::DNSClass::ANY => DNSClass::ANY,
            trust_dns_types::DNSClass::OPT(v) => DNSClass::OPT(v),
        }
    }
}


#[derive(Deserialize, Serialize)]
pub struct Record {
    #[serde(rename = "Name")]
    pub name: String,
    //#[serde(rename = "Type")]
    //rr_type: RecordType,
    #[serde(rename = "Class")]
    pub dns_class: DNSClass,
    #[serde(rename = "TTL")]
    pub ttl: u32,
    #[serde(flatten)]
    pub rdata: RData,
}

impl From<trust_dns_types::Record> for Record {
    fn from(record: trust_dns_types::Record) -> Record {
        Record {
            name: record.name().to_utf8(),
            //rr_type: record.rr_type().into(),
            dns_class: record.dns_class().into(),
            ttl: record.ttl(),
            rdata: record.into_data().into(),
        }
    }
}
